/*
  ex41.cu
  This is an example for testing the stream execution of CUDA kernels and memory transfer, simplifying the tests from CUDA-SDK simpleMultiCopy and simpleStreams.
  This software contains source code provided by NVIDIA Corporation.
 */

#include <stdio.h>
#include <stdlib.h>

__global__ void mykernel1(int *cudaA, const int numIterations, const int dataPerThread)
{
  int index = blockIdx.x * blockDim.x + threadIdx.x;
  int k;

  for( int i = 0; i < numIterations; i++ )
  {
    k = index * dataPerThread;
    for( int j = 0; j < dataPerThread; j++ )
      cudaA[k++] += index;
  }
}

__global__ void mykernel2(int *cudaA, const int numIterations, const int dataPerThread)
{
  int index = blockIdx.x * blockDim.x + threadIdx.x;
  int k;

  for( int i = 0; i < numIterations; i++ )
  {
    k = index * dataPerThread;
    for( int j = 0; j < dataPerThread; j++ )
      cudaA[k++] += 1;
  }
}

__global__ void kernel1_stream(int *cudaA, const int numIterations, 
    const int dataPerThread, const int streamNum)
{
  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  int index = gridDim.x * blockDim.x * streamNum + blockIdx.x * blockDim.x + threadIdx.x;
  int k;

  for( int i = 0; i < numIterations; i++ )
  {
    k = idx * dataPerThread;
    for( int j = 0; j < dataPerThread; j++ )
      cudaA[k++] += index;
  }
}

__global__ void kernel2_stream(int *cudaA, const int numIterations, 
    const int dataPerThread, const int streamNum)
{
  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  int k;

    k = idx * dataPerThread;
    for( int j = 0; j < dataPerThread; j++ )
      cudaA[k++] /= numIterations;
}



void check_data(const int *data, const int num, const int *correct)
{
  int i;
  printf( "Checking results......\n" );
  for( i = 0; i < num; i++ )
  {
    if( data[i] != correct[i] )
    {
      printf( "Wrong!!! data[%d]: %d. Correct value correct[%d]: %d\n\n", 
	  i, data[i], i, correct[i] );
      return;
    }
  }
  printf( "No mistakes in calculations\n\n" );
}

void runTest(int argc, char *argv[])
{
  int i, j, k;
  int numStreams = 8;

  if( argc > 1 ) numStreams = atoi(argv[1]);

  int numPoints = 8 * 4096 * 4096;
  int numThreads = 64;
  int numBlocks = 8;
  int numIterations = 5;

  int *correct;

  int dataPerThread = numPoints / numThreads / numBlocks / numStreams;

  cudaSetDevice(1);

  cudaEvent_t start, stop;
  cudaEventCreate(&start);
  cudaEventCreate(&stop);

  float memcpyTime = 0;
  float kernel1Time = 0;
  float kernel2Time = 0;
  float streamTotalTime = 0;

  int *a;
  int *cudaA;
  int streamSize = numPoints / numStreams;

  cudaStream_t stream[numStreams];

  correct = (int *)calloc( numPoints, sizeof(int) );
  /* First calculate the correct values that we are expecting in CPU */
  k = 0;
  for( i = 0; i < numThreads * numBlocks * numStreams; i++ )
  {
    for( j = 0; j < dataPerThread; j++ )
    {
      correct[k++] = i /* numIterations + numIterations*/;
    }
  }

  cudaMallocHost( (void **)&a, sizeof(int) * numPoints );
  cudaMalloc( (void **)&cudaA, sizeof(int) * numPoints );
  cudaMemset( cudaA, 0, sizeof(float) * numPoints );
  
  dim3 grid(numBlocks, 1);
  dim3 block(numThreads, 1);

  /* Measure the time of memcpy and kernel execution separately */
  cudaEventRecord(start, 0);

  for( i = 0; i < numStreams; i++ )
    kernel1_stream<<< grid, block >>>(cudaA + i * streamSize, numIterations,
	dataPerThread, i);

  cudaEventRecord(stop, 0);
  cudaEventSynchronize(stop);

  cudaEventElapsedTime( &kernel1Time, start, stop );

  printf( "Kernel1 execution time:\t%f (ms)\n", kernel1Time );


  /* Timing kernel2 */
  cudaEventRecord(start, 0);

  for( i = 0; i < numStreams; i++ )
    kernel2_stream<<< grid, block >>>(cudaA + i * streamSize, numIterations, dataPerThread, i);

  cudaEventRecord(stop, 0);
  cudaEventSynchronize(stop);

  cudaEventElapsedTime( &kernel2Time, start, stop );

  printf( "Kernel2 execution time:\t%f (ms)\n", kernel2Time );

  /* Timing memory transfer */
  cudaEventRecord(start, 0);

  cudaMemcpy( a, cudaA, sizeof(int) * numPoints, cudaMemcpyDeviceToHost );
  
  cudaEventRecord(stop, 0);
  cudaEventSynchronize(stop);

  cudaEventElapsedTime( &memcpyTime, start, stop );

  printf( "Memcpy time:\t\t%f (ms)\n", memcpyTime );

  printf( "Total time: \t\t%f (ms)\n", 
      kernel1Time + kernel2Time + memcpyTime );

  check_data(a, numPoints, correct );

  cudaMemset( cudaA, 0, sizeof(int) * numPoints );
  memset( a, 0, sizeof(int) * numPoints );

  /* Measure the time with Cuda streams. */
  cudaEventRecord(start, 0);

  for( i = 0; i < numStreams; i++ )
    cudaStreamCreate(&stream[i]);

  for( i = 0; i < numStreams; i++ )
    kernel1_stream<<< grid, block, 0, stream[i] >>>( cudaA + i * streamSize, 
	numIterations, dataPerThread, i );

  for( i = 0; i < numStreams; i++ )
    kernel2_stream<<< grid, block, 0, stream[i] >>>( cudaA + i * streamSize, 
	numIterations, dataPerThread, i );

  for( i = 0; i < numStreams; i++ )
    cudaMemcpyAsync( a + i * streamSize, cudaA + i * streamSize, sizeof(int) * streamSize,
       cudaMemcpyDeviceToHost, stream[i] );

  cudaEventRecord(stop, 0);
  cudaEventSynchronize(stop);

  cudaEventElapsedTime( &streamTotalTime, start, stop );

  for( i = 0; i < numStreams; i++ )
    cudaStreamDestroy(stream[i]);

  printf( "Stream execution time:\t%f (ms)\n", streamTotalTime );

  check_data( a, numPoints, correct );

  cudaFreeHost(a);
  cudaFree(cudaA);
  free(correct);
}


int main(int argc, char *argv[])
{
  runTest(argc, argv);

  return 0;
}
