/*
  ex43.cu
  This is an example showing how to avoid branches in 
  kernel with simple maths operations.
 */
#include <stdio.h>

/* A kernel that limits the values by 255, assuming positive integers */
__global__ void branch1_kernel(int *input, int *output)
{
  int index = blockIdx.x * blockDim.x + threadIdx.x;

  if( input[index] <= 255 ) output[index] = input[index];
  else output[index] = 255;
}
/* Math version of the same kernel */
__global__ void math1_kernel(int *input, int *output)
{
  int index = blockIdx.x * blockDim.x + threadIdx.x;

  output[index] = min(input[index], 255);
}

/* An artificially chosen kernel to demonstrate the extreme case (sort of) for warp divergence */
__global__ void branch2_kernel(int *output)
{
  int index = blockIdx.x * blockDim.x + threadIdx.x;

  if( threadIdx.x % 32 == 0 )
    output[index] = 0;
  else if( threadIdx.x % 32 == 1 )
    output[index] = 1;
  else if( threadIdx.x % 32 == 2 )
    output[index] = 2;
  else if( threadIdx.x % 32 == 3 )
    output[index] = 3;
  else if( threadIdx.x % 32 == 4 )
    output[index] = 4;
  else 
    output[index] = threadIdx.x % 32;
}
/* Math version of the same kernel */
__global__ void math2_kernel(int *output)
{
  int index = blockIdx.x * blockDim.x + threadIdx.x;

  output[index] = threadIdx.x % 32;
}

int main(int argc, char *argv[])
{
  int i, j;

  int nthreads = 512;
  int nblocks = 2048;

  int arrSize = nthreads * nblocks;

  /* Set the inputs */
  int *input1;
  int *cudaInput1;
  int *cudaOutput1, *cudaOutput2;

  /* Allocate the memories */
  input1 = (int *)malloc( arrSize * sizeof(int) );

  cudaMalloc( (void **)&cudaInput1, arrSize * sizeof(int) );
  cudaMalloc( (void **)&cudaOutput1, arrSize * sizeof(int) );
  cudaMalloc( (void **)&cudaOutput2, arrSize * sizeof(int) );

  srand(201210);
  /* Now initilize the arrays, first one can be just any below 256 */
  for( i = 0; i < arrSize; i++ )
    input1[i] = rand() / 256;

  /* Now set the possible warp divergence, it is setup in such a way that you get warp divergence in the first thread of the warp */
  for( i = 0; i < nblocks; i++ )
  {
    for( j = 0; j < 16; j+=32 )
    {
      input1[i*nthreads+j] = 500;
    }
  }

  /* Now copy the memory */
  cudaMemcpy( cudaInput1, input1, arrSize * sizeof(int), cudaMemcpyHostToDevice );

  branch1_kernel<<< nblocks, nthreads >>>(cudaInput1, cudaOutput1);
  
  math1_kernel<<< nblocks, nthreads >>>(cudaInput1, cudaOutput1);

  branch2_kernel<<< nblocks, nthreads >>>(cudaOutput2);

  math2_kernel<<< nblocks, nthreads >>>(cudaOutput2);

  free(input1);
  cudaFree(cudaInput1);
  cudaFree(cudaOutput1);
  cudaFree(cudaOutput2);

  return 0;
}
